// Copyright (C) 2019  Wilko Manger
//
// This file is part of matrix-dart-sdk.
//
// matrix-dart-sdk is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// matrix-dart-sdk is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
//
// You should have received a copy of the GNU Affero General Public License
// along with matrix-dart-sdk.  If not, see <https://www.gnu.org/licenses/>.

import 'package:matrix_sdk/src/event/room/room_event.dart';
import 'package:matrix_sdk/src/event/room/state/state_event.dart';
import 'package:matrix_sdk/src/identifier.dart';
import 'package:matrix_sdk/src/user.dart';
import 'package:meta/meta.dart';

const String _roomMemberType = 'm.room.member';

abstract class MemberChangeEvent extends StateEvent {
  static const String type = _roomMemberType;

  @override
  final String matrixType = type;

  @override
  final MemberChange content;

  MemberChangeEvent(
      this.content, RoomEventArgs args)
      : super(args, stateKey: content.subject.id.toString());

  factory MemberChangeEvent.fromJson(RoomEventArgs args, String stateKey, Map<String, dynamic> json) {
    final membershipType = json['content']['membership'];
    final subjectId = UserId(stateKey);

    final prevContent = json.containsKey('unsigned')
                          && json['unsigned'] != null
                       ? json['unsigned']['prev_content'] : null;

    final prevDisplayName = prevContent != null
                          ? prevContent['displayname'] : null;

    final displayName = json['content']['displayname'] ?? prevDisplayName;

    var prevAvatarUrl = prevContent != null
                        ? prevContent['avatar_url'] : null;
    if (prevAvatarUrl != null) {
      prevAvatarUrl = Uri.parse(prevAvatarUrl);
    }

    var avatarUrl = json['content']['avatar_url'] ?? prevAvatarUrl;

    if (avatarUrl != null && avatarUrl is String) {
      avatarUrl = Uri.parse(avatarUrl);
    }

    final subject = User(
      id: subjectId,
      state: UserState(
        roomId: args.room.id,
        since: args.time,
        displayName: displayName,
        avatarUrl: avatarUrl
      )
    );

    if (prevContent != null
      && prevContent['membership'] == 'join'
      && membershipType == 'join') {
      if (displayName != prevDisplayName) {
        return DisplayNameChangeEvent(
          subject,
          args,
          oldSubjectName: prevDisplayName
        );
      }

      if (avatarUrl != prevAvatarUrl) {
        return AvatarUrlChangeEvent(
          subject,
          args,
          oldSubjectAvatarUrl: prevAvatarUrl,
        );
      }
    }

    switch (membershipType) {
      case JoinEvent.type:
        return JoinEvent(subject, args);
      case LeaveEvent.type:
        return LeaveEvent(subject, args);
      case BanEvent.type:
        return BanEvent(subject, args);
      case InviteEvent.type:
        return InviteEvent(subject, args);
      case KnockEvent.type:
      default:
        return KnockEvent(subject, args);
    }
  }
}

class MemberChange extends StateEventContent {
  @override
  final String matrixType = _roomMemberType;

  @override
  final String stateKey;

  /// The [User] of who has joined, left, was banned, etc.
  final User subject;

  @protected
  final String membershipType;

  MemberChange(
    {
      @required this.subject,
      @required this.membershipType,
    })
    : stateKey = subject.id.toString();

  @override
  Map<String, dynamic> toJson() => super.toJson()
    ..addAll({
      'membership': membershipType,
      'displayname': subject.name,
      'avatar_url': subject.avatarUrl.toString()
    });
}

class JoinEvent extends MemberChangeEvent {
  static const String type = 'join';

  JoinEvent(
      User subject, RoomEventArgs args)
      : super(MemberChange(
        subject: subject,
        membershipType: type
      ), args);
}

class LeaveEvent extends MemberChangeEvent {
  static const String type = 'leave';

  LeaveEvent(
    User subject, RoomEventArgs args)
      : super(MemberChange(
        subject: subject,
        membershipType: type
      ),
      args);
}

class BanEvent extends MemberChangeEvent {
  static const String type = 'ban';

  BanEvent(User subject, RoomEventArgs args)
    : super(MemberChange(
      subject: subject,
      membershipType: type
    ),
    args);
}

// TODO: Add RoomState
class InviteEvent extends MemberChangeEvent {
  static const String type = 'invite';

  InviteEvent(User subject, RoomEventArgs args)
    : super(MemberChange(
      subject: subject,
      membershipType: type
    ), args);
}

class KnockEvent extends MemberChangeEvent {
  static const String type = 'knock';

  KnockEvent(User subject, RoomEventArgs args)
      : super(MemberChange(
      subject: subject,
      membershipType: type,
  ), args);
}

class DisplayNameChangeEvent extends JoinEvent {
  final String oldSubjectName;
  String get newSubjectName => content.subject.name;

  DisplayNameChangeEvent(
    User subject,
    RoomEventArgs args,
    {@required this.oldSubjectName}
  ) : super(subject, args);

  @override
  Map<String, dynamic> toJson() => super.toJson()
    ..addAll({
      'unsigned': {
        'prev_content': {
          'membership': 'join',
          'displayname': oldSubjectName,
          'avatar_url': content.subject.avatarUrl.toString()
        }
      }
    });
}

class AvatarUrlChangeEvent extends JoinEvent {
  final Uri oldSubjectAvatarUrl;
  Uri get newSubjectAvatarUrl => content.subject.avatarUrl;

  AvatarUrlChangeEvent(
    User subject,
    RoomEventArgs args,
    {@required this.oldSubjectAvatarUrl}
  ) : super(subject, args);

  @override
  Map<String, dynamic> toJson() => super.toJson()
    ..addAll({
      'unsigned': {
        'prev_content': {
          'membership': 'join',
          'displayname': content.subject.name,
          'avatar_url': oldSubjectAvatarUrl.toString()
        }
      }
    });
}