// Copyright (C) 2019  Wilko Manger
//
// This file is part of matrix-dart-sdk.
//
// matrix-dart-sdk is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// matrix-dart-sdk is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
//
// You should have received a copy of the GNU Affero General Public License
// along with matrix-dart-sdk.  If not, see <https://www.gnu.org/licenses/>.

import 'dart:async';
import 'dart:collection';

import 'package:matrix_sdk/src/storables.dart';
import 'package:matrix_sdk/src/identifier.dart';
import 'package:matrix_sdk/src/store.dart';
import 'package:matrix_sdk/src/user.dart';
import 'package:meta/meta.dart';

class Members implements Storables<UserId, User> {

  @override
  final Store store;

  final _memory = SplayTreeSet<User>();
  final RoomId _roomId;

  Members(this._roomId, {this.store});

  @override
  @protected
  Future<User> retrieve(UserId id) => store?.getMember(_roomId, id);

  @override
  @protected
  Stream<User> retrieveAll() => store?.getMembers(_roomId);

  @override
  @protected
  Future<void> saveAll(Iterable<User> users) => store?.addMembers(_roomId, users);

  @override
  @protected
  Stream<User> retrieveUpTo(int count, {@required User from})
  => store?.getMembers(_roomId);

  @override
  FutureOr<User> operator [](UserId id) => get(id);

  @override
  Future<void> addAll(Iterable<User> values) async {
    _memory.addAll(values);
    await saveAll(values);
  }

  @override
  /// Returns all states of all users.
  Stream<User> all() async* {
    for (final item in _memory) {
      yield item;
    }

    // Load all states to memory
    await for (final item in retrieveAll()) {
      // If the state was not already in memory, yield it
      final added = _memory.add(item);
      if (added) {
        yield item;
      }
    }
  }

  @override
  Future<void> remove(UserId id) async
    => _memory.removeWhere((user) => user.id == id);

  @override
  void removeInMemory(UserId id) => null;

  @override
  Future<void> removeInStore(UserId id) => null;

  @override
  Stream<User> retrieveRemote(int count, {User from}) => null;

  @override
  Stream<User> upTo(int count) async* {
    var i = 0;
    var item;

    final uniqueUsers = SplayTreeSet<User>.of(_memory,
      (a, b) => a.id.toString().compareTo(b.id.toString())
    );
    final memory = uniqueUsers.map((user) => get(user.id));

    for (item in memory) {
      if (item is Future) {
        item = await item;
      }

      yield item;
      i++;

      if (i >= count) {
        return;
      }
    }

    await for (item in retrieveUpTo(count - i, from: item)) {
      // If the event was not already in memory, yield it
      final added = _memory.add(item);
      if (added) {
        yield item;
        i++;
      }

      if (i >= count) {
        return;
      }
    }
  }

  FutureOr<User> get(UserId id, {DateTime at}) {
    if (at == null) {
      at = DateTime.now();
    }

    final users = _memory
        .where((user) => user.id == id)
        .toList(growable: false);

    for (var i = 0; i < users.length; i++) {
      final current = users[i];
      final next = i + 1 < users.length ? users[i + 1] : null;

      final currentIsBefore =
        current.state.since.millisecondsSinceEpoch <= at.millisecondsSinceEpoch;

      final nextIsAfterOrNull = next == null
        || next.state.since.millisecondsSinceEpoch > at.millisecondsSinceEpoch;

      if (currentIsBefore && nextIsAfterOrNull) {
        return current;
      }
    }

    // If not in memory, return from store
    return store?.getMember(_roomId, id, at: at);
  }

  @override
  Iterable<User> sort(Iterable<User> items) => items;
}
